from typing import Any, Callable

from aiohttp import web

from bolinette.core import Cache, meta
from bolinette.core.injection import Injection, init_method
from bolinette.core.injection.resolver import ArgResolverOptions
from bolinette.core.types import Type
from bolinette.core.utils import AttributeUtils
from bolinette.web import Controller
from bolinette.web.controller import ControllerMeta
from bolinette.web.middleware import Middleware, MiddlewareBag
from bolinette.web.route import RouteBucket, RouteProps


class WebResources:
    def __init__(self) -> None:
        self.web_app = web.Application()

    @init_method
    def _init_ctrls(self, cache: Cache, inject: Injection) -> None:
        routes: list[web.RouteDef] = []
        for ctrl_cls in cache.get(ControllerMeta, hint=type[Controller]):
            ctrl_meta = meta.get(ctrl_cls, ControllerMeta)
            attr: Any
            for attr in AttributeUtils.get_cls_attrs(ctrl_cls).values():
                if not meta.has(attr, RouteBucket):
                    continue
                bucket = meta.get(attr, RouteBucket)
                for props in bucket.routes:
                    path = props.anon_path
                    if not path:
                        path = ctrl_meta.path
                    elif not path.startswith("/"):
                        path = f"{ctrl_meta.path}/{path}"
                    if path and not path.startswith("/"):
                        path = f"/{path}"
                    match props.method:
                        case "GET":
                            handler = web.get
                        case "POST":
                            handler = web.post
                        case "PUT":
                            handler = web.put
                        case "PATCH":
                            handler = web.patch
                        case "DELETE":
                            handler = web.delete
                        case _:
                            raise NotImplementedError()
                    routes.append(handler(path, RouteHandler(inject, Type(ctrl_cls), props, attr).handle))
        self.web_app.add_routes(routes)


class RouteHandler:
    __slots__ = ("inject", "ctrl_t", "props", "func")

    def __init__(
        self,
        inject: Injection,
        ctrl_t: Type[Controller],
        props: RouteProps,
        func: Callable[..., Any],
    ) -> None:
        self.inject = inject
        self.ctrl_t = ctrl_t
        self.props = props
        self.func = func

    @staticmethod
    def prepare_session(inject: Injection, request: web.Request) -> None:
        inject.add(web.Request, "scoped", instance=request)

    async def handle(self, request: web.Request) -> web.Response:
        try:
            scoped_inject = self.inject.get_scoped_session()
            self.prepare_session(scoped_inject, request)
            mdlws = self.collect_middlewares(scoped_inject)
            return await self.middleware_chain(request, scoped_inject, mdlws, 0)
        except BaseException as e:
            raise e

    async def middleware_chain(
        self,
        request: web.Request,
        scoped: Injection,
        mdlws: list[Middleware[Any]],
        index: int,
    ) -> web.Response:
        async def _next_handle() -> web.Response:
            return await self.middleware_chain(request, scoped, mdlws, index + 1)

        if index >= len(mdlws):
            ctrl = scoped.instanciate(self.ctrl_t.cls)
            return await scoped.call(
                self.func,
                args=[ctrl],
                additional_resolvers=[RouteParamArgResolver(request, self.props)],
            )
        else:
            mdlw = mdlws[index]
            return await scoped.call(mdlw.handle, args=[_next_handle])

    def collect_middlewares(self, scoped: Injection) -> list[Middleware[Any]]:
        bags: list[MiddlewareBag] = []
        if meta.has(self.ctrl_t.cls, MiddlewareBag):
            bags.append(meta.get(self.ctrl_t.cls, MiddlewareBag))
        if meta.has(self.func, MiddlewareBag):
            bags.append(meta.get(self.func, MiddlewareBag))
        mdlws: dict[Type[Middleware[...]], Middleware[...]] = {}
        for bag in bags:
            for t, mdlw_meta in reversed(bag.added.items()):
                mdlw = scoped.instanciate(t.cls)
                mdlw.options(*mdlw_meta.args, **mdlw_meta.kwargs)
                mdlws[t] = mdlw
            for t in bag.removed:
                if t in mdlws:
                    del mdlws[t]
        return list(mdlws.values())


class RouteParamArgResolver:
    def __init__(self, request: web.Request, props: RouteProps) -> None:
        self.request = request
        self.props = props

    def supports(self, options: ArgResolverOptions):
        return (
            options.name in self.props.func_to_url_args
            and self.props.func_to_url_args[options.name] in self.request.match_info
        )

    def resolve(self, options: ArgResolverOptions) -> tuple[str, Any]:
        value = self.request.match_info[self.props.func_to_url_args[options.name]]
        t = options.t
        match t.cls:
            case cls if cls is int:
                value = int(value)
            case cls if cls is float:
                value = float(value)
            case cls if cls is bool:
                value = bool(value)
            case cls if cls is not str:
                raise Exception()  # TODO: Lever une vraie exception
            case _:
                pass
        return (options.name, value)



from bolinette.web.version import __version__
from bolinette.web.route import (
    route as route,
    get as get,
    post as post,
    put as put,
    patch as patch,
    delete as delete,
)
from bolinette.web.controller import Controller as Controller, controller as controller
from bolinette.web.middleware import with_middleware as with_middleware, without_middleware as without_middleware
from bolinette.web.resources import WebResources as WebResources



from typing import Callable, TypeVar

from bolinette.core import Cache, __user_cache__, meta


class Controller:
    pass


class ControllerMeta:
    def __init__(self, path: str) -> None:
        self.path = path


TCtrl = TypeVar("TCtrl", bound=Controller)


def controller(path: str, /, *, cache: Cache | None = None) -> Callable[[type[TCtrl]], type[TCtrl]]:
    def decorator(cls: type[TCtrl]) -> type[TCtrl]:
        meta.set(cls, ControllerMeta(path))
        (cache or __user_cache__).add(ControllerMeta, cls)
        return cls

    return decorator



from typing import Any, Awaitable, Callable, ParamSpec, Protocol, TypeVar

from aiohttp import web

from bolinette.core import meta
from bolinette.core.types import Type
from bolinette.web import Controller

MdlwInitP = ParamSpec("MdlwInitP")
CtrlT = TypeVar("CtrlT", bound=Controller | Callable[..., Any])


class Middleware(Protocol[MdlwInitP]):
    def options(self, *args: MdlwInitP.args, **kwargs: MdlwInitP.kwargs) -> None:
        ...

    async def handle(self, next: Callable[[], Awaitable[web.Response]]) -> web.Response:
        ...


class MiddlewareMeta:
    def __init__(self, args: tuple[Any, ...], kwargs: dict[str, Any]) -> None:
        self.args = args
        self.kwargs = kwargs


class MiddlewareBag:
    def __init__(self) -> None:
        self.added: dict[Type[Middleware[...]], MiddlewareMeta] = {}
        self.removed: list[Type[Middleware[...]]] = []


def with_middleware(
    middleware: type[Middleware[MdlwInitP]],
    *args: MdlwInitP.args,
    **kwargs: MdlwInitP.kwargs,
) -> Callable[[CtrlT], CtrlT]:
    def decorator(func: CtrlT) -> CtrlT:
        if not meta.has(func, MiddlewareBag):
            bag = MiddlewareBag()
            meta.set(func, bag)
        else:
            bag = meta.get(func, MiddlewareBag)
        bag.added[Type(middleware)] = MiddlewareMeta(args, kwargs)
        return func

    return decorator


def without_middleware(middleware: type[Middleware[...]]) -> Callable[[CtrlT], CtrlT]:
    def decorator(func: CtrlT) -> CtrlT:
        if not meta.has(func, MiddlewareBag):
            bag = MiddlewareBag()
            meta.set(func, bag)
        else:
            bag = meta.get(func, MiddlewareBag)
        bag.removed.append(Type(middleware))
        return func

    return decorator



import re
from typing import Callable, Literal, ParamSpec, TypeVar

from bolinette.core import meta

FuncP = ParamSpec("FuncP")
FuncT = TypeVar("FuncT")
HttpMethod = Literal["GET", "POST", "PUT", "PATCH", "DELETE"]
path_re = re.compile(r"\{([^}]+)\}")


class RouteProps:
    def __init__(self, method: HttpMethod, path: str) -> None:
        self.method = method
        self.real_path = path
        self.anon_path, self.url_to_func_args, self.func_to_url_args = self.build_anon_path(path)

    @staticmethod
    def build_anon_path(path: str) -> tuple[str, dict[str, str], dict[str, str]]:
        new_path: list[str] = []
        last_pos = 0
        anon_to_param: dict[str, str] = {}
        param_to_anon: dict[str, str] = {}
        for index, match in enumerate(path_re.finditer(path)):
            start, end = match.span()
            new_path.append(path[last_pos:start])
            anon_param = f"_p{index}"
            anon_to_param[anon_param] = match.group(1)
            param_to_anon[match.group(1)] = anon_param
            new_path.append(f"{{{anon_param}}}")
            last_pos = end
        new_path.append(path[last_pos:])
        return "".join(new_path), anon_to_param, param_to_anon


class RouteBucket:
    def __init__(self, name: str) -> None:
        self.name = name
        self.routes: list[RouteProps] = []

    def add(self, route: RouteProps) -> None:
        self.routes.append(route)

    def __getitem__(self, index: int) -> RouteProps:
        return self.routes[index]


def route(method: HttpMethod, path: str, /) -> Callable[[Callable[FuncP, FuncT]], Callable[FuncP, FuncT]]:
    def decorator(func: Callable[FuncP, FuncT]) -> Callable[FuncP, FuncT]:
        if not meta.has(func, RouteBucket):
            bucket = RouteBucket(func.__name__)
            meta.set(func, bucket)
        else:
            bucket = meta.get(func, RouteBucket)
        bucket.add(RouteProps(method, path))
        return func

    return decorator


def get(path: str, /) -> Callable[[Callable[FuncP, FuncT]], Callable[FuncP, FuncT]]:
    return route("GET", path)


def post(path: str, /) -> Callable[[Callable[FuncP, FuncT]], Callable[FuncP, FuncT]]:
    return route("POST", path)


def put(path: str, /) -> Callable[[Callable[FuncP, FuncT]], Callable[FuncP, FuncT]]:
    return route("PUT", path)


def patch(path: str, /) -> Callable[[Callable[FuncP, FuncT]], Callable[FuncP, FuncT]]:
    return route("PATCH", path)


def delete(path: str, /) -> Callable[[Callable[FuncP, FuncT]], Callable[FuncP, FuncT]]:
    return route("DELETE", path)
